//
// Created by Attila Kákay on 12/15/21.
//
#include <stdio.h>
#include <stdlib.h>
#include "fempreproc.h"
#include "akmag.h"
#include "utils.h"
#include "bnd2d.h"
#include "galerkin.h"

int OperatorsFromMesh(double *xyz, int *ijk, int nx, int nelx, int dim, double Aex, double Msat,
                      double dtet, int *bnlist, int *bnx, double *nv, double *area,
                      CSRSprMat *xc, CSRSprMat *poiss, CSRSprMat *lap,
                      CSRSprMat *div_x, CSRSprMat *div_y,
                      CSRSprMat *grad_x, CSRSprMat *grad_y,
//                      CSRSprMat *div_x, CSRSprMat *div_y, CSRSprMat *div_z,
//                      CSRSprMat *grad_x, CSRSprMat *grad_y, CSRSprMat *grad_z,
                      double *pang, int *belm)
{
    int bnelx, err, eld, nrg=1, i,j;

    AKElm elm, belem;
    OperatorMat OpMat;

    N_AKElm(&elm, nelx, nx, dim);
    // Here I need to set elm.xyz from sample.xyz
    // elm.ijk from sample.mesh
    elm.xyz = xyz;
    for(i=0; i<nelx; i++){
        elm.ijk[0][i] = ijk[3*i];
        elm.ijk[1][i] = ijk[3*i+1];
        elm.ijk[2][i] = ijk[3*i+2];
        elm.ijk[3][i] = 1;
    }

    clean_mesh(&elm);

    if(nx != elm.nx) {
        printf("Found ghost nodes!\n");
        fflush(stdout);
        exit(0);
    }


    elm.nrg = nrg;

    elm.dtet = dtet;
    elm.params = (MatParams *) malloc(nrg * sizeof(MatParams));
    elm.params[0].A = Aex;
    elm.params[0].Ms = Msat;
    elm.params[0].Bs = Msat*mu0;


    //dim = elm.dim;
    //eld = dim + 1;
//    elm.params = sample.mat

    printf("Making the inverse node-element list.\n");
    MakeInvNodeElmList(&elm);
    printf("Making the adjacent node list.\n");
    MakeAdjacentNodeList(&elm);

    // BEGINNING: Boundary elements, nodes
    printf("Counting the number of boundary elements and nodes.\n");
    bndprocessing(&elm, &belem, &bnelx);
    printf("The number of boundary elements: %d\n",bnelx);
    printf("The number of boundary nodes: %d\n",elm.bnx);

    *bnx = elm.bnx;

    printf("Computing the static Dense Matrix.\n");
//    if(dim == 2) ComputeDenseMatrix2D(&elm,belem,&OpMat);
//  else ComputeDenseMatrix3D(&fem3D,belm,&OpMat);
    printf("DONE.\n");
// END: Boundary elements, nodes

// BEGIN: Computing operators/matrices
    printf("Computing shape functions and matrices.\n");
    ShapeFGrad(&elm);
    GalerkinInteractions(&elm,&OpMat);
    OpMat.xc_expanded =  expand_dimension(OpMat.xct);
    ris2csrDirect(&OpMat.xc_expanded,xc);
    ris2csrDirect(&OpMat.poiss,poiss);
    ris2csrDirect(&OpMat.lap,lap);
    ris2csrDirect(&OpMat.ldx,div_x);
    ris2csrDirect(&OpMat.ldy,div_y);
//    if(dim == 3) ris2csrDirect(&OpMat.ldz,div_z);
    ris2csrDirect(&OpMat.phix,grad_x);
    ris2csrDirect(&OpMat.phiy,grad_y);
//    if(dim == 3) ris2csrDirect(&OpMat.phiz,grad_z);
    printf("DONE.\n");
//#pragma parallel for
    for(i=0;i<elm.bnx;i++) {
        bnlist[i] = elm.bnlist[i];
        pang[i] = elm.sub[i];
        belm[2*i] = belem.ijk[0][i];
        belm[2*i+1] = belem.ijk[1][i];
    }
    for (i=0;i<elm.nx;i++) area[i] = elm.vol_n[i];
    for (j=0;j<dim*bnelx;j++) nv[j] = elm.nv[j];

    return(1);
}

