import pygmsh


def tube_cross_section(r, R, lc=5):
    with pygmsh.occ.Geometry() as geom:
        geom.characteristic_length_min = lc
        geom.characteristic_length_max = lc
        disk1 = geom.add_disk([0, 0.0], R)
        disk2 = geom.add_disk([0, 0.0], r)
        geom.boolean_difference(disk1, disk2)
        mesh = geom.generate_mesh()
    return mesh


def round_wire_cross_section(R, lc=5):
    with pygmsh.occ.Geometry() as geom:
        geom.characteristic_length_min = lc
        geom.characteristic_length_max = lc
        disk1 = geom.add_disk([0, 0.0], R)
        mesh = geom.generate_mesh()
    return mesh


def polygonal_cross_section(points, lc=5):
    with pygmsh.geo.Geometry() as geom:
        geom.add_polygon(points, mesh_size=lc)
        mesh = geom.generate_mesh()
    return mesh


def rectangle_cross_section(a, b, lc_a=5, lc_b=5):
    """
    Creates a mesh with rectangular cross section.

    Parameters
    ----------
    a : length
    b : width
    lc_a : characteristic mesh length long a
    lc_b : characteristic mesh length long b

    """
    with pygmsh.geo.Geometry() as geom:
        nx = int(a / lc_a)
        ny = int(b / lc_b)
        p = geom.add_point([-a / 2, -b / 2, 0.0], mesh_size=lc_a)
        _, l, _ = geom.extrude(p, translation_axis=[a, 0, 0], num_layers=nx)
        geom.extrude(l, translation_axis=[0, b, 0], num_layers=ny)
        mesh = geom.generate_mesh()
    return mesh


def rectangle(a, b, lc=5):
    with pygmsh.geo.Geometry() as geom:
        geom.characteristic_length_min = lc
        geom.characteristic_length_max = lc
        geom.add_rectangle(0.0, 2.0, 0.0, 1.0, 0.0, 0.1)
        mesh = geom.generate_mesh()
    return mesh
